"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.Memory = void 0;
exports.instantiate = instantiate;
const {
  Buffer
} = require('buffer');
const {
  nanoid
} = require('nanoid/non-secure');
const Webassembly = require('./NativeWebassembly').default;
class Memory {
  constructor(params) {
    this.__initial = params === null || params === void 0 ? void 0 : params.initial;
  }
}
exports.Memory = Memory;
const DEFAULT_STACK_SIZE_IN_BYTES = 8192;
const DEFAULT_MEMORY = new Memory({
  initial: DEFAULT_STACK_SIZE_IN_BYTES
});
const getScopedFunctions = importsMap => {
  if (!importsMap) return [];
  return Object.entries(importsMap).flatMap(_ref => {
    let [scope, imports] = _ref;
    return Object.keys(imports).map(functionName => ({
      scope,
      functionName
    }));
  });
};

// https://developer.mozilla.org/en-US/docs/WebAssembly/JavaScript_interface/instantiate
async function instantiate(bufferSource) {
  let maybeImportObject = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : undefined;
  const iid = nanoid();
  const importObject = maybeImportObject || {};
  const {
    env: maybeEnv,
    ...extras
  } = importObject;
  const memory = (maybeEnv === null || maybeEnv === void 0 ? void 0 : maybeEnv.memory) || DEFAULT_MEMORY;
  const stackSizeInBytes = (memory === null || memory === void 0 ? void 0 : memory.__initial) ?? DEFAULT_STACK_SIZE_IN_BYTES;
  const scopedFunctions = getScopedFunctions(extras);
  const rawFunctions = scopedFunctions.map(_ref2 => {
    let {
      functionName
    } = _ref2;
    return functionName;
  });
  const rawFunctionScopes = scopedFunctions.map(_ref3 => {
    let {
      scope
    } = _ref3;
    return scope;
  });
  const instanceResult = Webassembly.instantiate({
    iid,
    bufferSource: Buffer.from(bufferSource).toString('base64'),
    stackSizeInBytes,
    rawFunctions,
    rawFunctionScopes
  });
  if (instanceResult !== 0) throw new Error('Failed to instantiate WebAssembly.');
  const exports = new Proxy({}, {
    get(_, func) {
      if (typeof func !== 'string') throw new Error(`Expected string, encountered ${typeof func}.`);
      return function () {
        for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
          args[_key] = arguments[_key];
        }
        const res = Webassembly.invoke({
          iid,
          func,
          args: args.map(e => e.toString())
        });
        if (!res.length) return undefined;
        const num = res.map(parseFloat);
        if (res.length !== 1) return num;
        return num[0];
      };
    }
  });
  return {
    instance: {
      exports
    }
  };
}
//# sourceMappingURL=index.js.map